import numpy as np
from numpy.typing import NDArray
import pytest
import scipy
import scipy.spatial

import delaunay_triangulation_and_its_dual_2d
import delaunay_triangulation_and_its_dual_2d.exceptions


def _sort_rows(array: NDArray) -> NDArray:
    return array[np.lexsort(array.T[::-1])]


def _sort_columns_of_each_row_then_sort_rows(array: NDArray) -> NDArray:
    array_columns_sorted = np.sort(array, axis=1)
    return _sort_rows(array=array_columns_sorted)


def _process_num_list_representing_a_polygon_for_equality_check(
    list_: list,
) -> list:
    if len(list_) == 1:
        return list_
    min_index = list_.index(min(list_))
    result = list_[min_index:] + list_[:min_index]
    if result[1] > result[-1]:
        result = [result[0]] + result[1:][::-1]
    return result


@pytest.mark.parametrize(
    "num_of_points, dimensions_of_points, result",
    [
        (100, 2, True),
        (3, 2, True),
        (3, 3, False),
        (3, 1, False),
        (2, 2, False),
        (1, 2, False),
    ],
)
def test_check_points_validity_method(
    num_of_points: int, dimensions_of_points: int, result: bool
):
    points = np.zeros((num_of_points, dimensions_of_points), dtype=np.float32)
    assert (
        delaunay_triangulation_and_its_dual_2d.Delaunay._check_points_validity(
            points=points
        )
        is result
    )


@pytest.mark.parametrize(
    "value",
    [
        -1.0,
        0.0,
        1.0,
        100000.0,
    ],
)
def test_duplicated_points_exist(value: float):
    identical_points = np.full((3, 2), fill_value=value, dtype=np.float32)
    assert (
        delaunay_triangulation_and_its_dual_2d.Delaunay._duplicated_points_exist(
            points=identical_points
        )
        is True
    )
    random_points = np.random.rand(3, 2)
    assert (
        delaunay_triangulation_and_its_dual_2d.Delaunay._duplicated_points_exist(
            points=random_points
        )
        is False
    )
    mixed_points = np.concatenate(
        [identical_points, random_points], axis=0, dtype=np.float32
    )
    assert (
        delaunay_triangulation_and_its_dual_2d.Delaunay._duplicated_points_exist(
            points=mixed_points
        )
        is True
    )


@pytest.mark.parametrize(
    "num_of_points",
    [
        3,
        4,
        100000,
    ],
)
def test_jitter_points_on_overlap(num_of_points: int):
    num_of_points_0 = num_of_points // 2
    num_of_points_1 = num_of_points - num_of_points_0
    points = np.concatenate(
        [np.zeros((num_of_points_0, 2)), np.ones((num_of_points_1, 2))],
        axis=0,
        dtype=np.float32,
    )
    points = delaunay_triangulation_and_its_dual_2d.Delaunay._jitter_points_on_overlap(
        points=points, rng=np.random.default_rng()
    )


@pytest.mark.parametrize(
    "num_of_points, num_of_iterations",
    [
        (3, 1),
        (3, 1000),
        (4, 1),
        (4, 1000),
        (100000, 1),
    ],
)
def test_voronoi_iteration(num_of_points: int, num_of_iterations: int):
    points = np.random.uniform(low=0.0, high=100.0, size=(num_of_points, 2))
    for _ in range(num_of_iterations):
        points = (
            delaunay_triangulation_and_its_dual_2d.Delaunay.voronoi_iteration(
                points=points, rng=np.random.default_rng()
            )
        )


def test_compute_voronoi_diagram():
    good_points_found = False
    for _ in range(1000):
        points = np.random.uniform(low=0.0, high=100.0, size=(100, 2))
        delaunay = delaunay_triangulation_and_its_dual_2d.Delaunay(
            points=points
        )
        try:
            delaunay.compute_voronoi_tessellation()
        except (
            delaunay_triangulation_and_its_dual_2d.DuplicatedCircumcentersError
        ) as e:
            print(str(e))
        else:
            good_points_found = True
            break
    if not good_points_found:
        raise ValueError("Cannot find good points in reasonable iterations")

    expected_outputs = scipy.spatial.Voronoi(points=points)
    delaunay = delaunay_triangulation_and_its_dual_2d.Delaunay(points=points)
    assert delaunay.voronoi_vertices.dtype == expected_outputs.vertices.dtype
    assert np.allclose(
        _sort_rows(array=delaunay.voronoi_vertices),
        _sort_rows(array=expected_outputs.vertices),
        atol=1e-6,
    )
    assert delaunay.ridge_points.dtype == expected_outputs.ridge_points.dtype
    assert np.array_equal(
        _sort_columns_of_each_row_then_sort_rows(array=delaunay.ridge_points),
        _sort_columns_of_each_row_then_sort_rows(
            array=expected_outputs.ridge_points
        ),
    )
    delaunay_ridge_vertices = np.array(delaunay.ridge_vertices, dtype=np.intc)
    expected_outputs_ridge_vertices = np.array(
        expected_outputs.ridge_vertices, dtype=np.intc
    )
    assert np.array_equal(
        _sort_columns_of_each_row_then_sort_rows(
            array=delaunay_ridge_vertices
        ),
        _sort_columns_of_each_row_then_sort_rows(
            array=expected_outputs_ridge_vertices
        ),
    )
    delaunay_regions = [
        _process_num_list_representing_a_polygon_for_equality_check(list_=r)
        for r in delaunay.regions
    ]
    delaunay_regions.sort(key=lambda x: tuple(x))
    expected_outputs_regions = [
        _process_num_list_representing_a_polygon_for_equality_check(list_=r)
        for r in expected_outputs.regions
        if len(r)
    ]
    expected_outputs_regions.sort(key=lambda x: tuple(x))
    assert delaunay_regions == expected_outputs_regions


def test_compute_voronoi_diagram_raises_duplicated_circumcenters_error():
    points = np.array(
        [[0, 0], [1, 0], [1, 1], [0, 1]],
        dtype=np.float32,
    )
    delaunay = delaunay_triangulation_and_its_dual_2d.Delaunay(points=points)
    with pytest.raises(
        delaunay_triangulation_and_its_dual_2d.exceptions.DuplicatedCircumcentersError
    ):
        delaunay.compute_voronoi_tessellation()



import cProfile
import pstats

import numpy as np

import delaunay_triangulation_and_its_dual_2d


rng = np.random.default_rng()
points = rng.uniform(low=0.0, high=1.0, size=20000).reshape(-1, 2)
i = delaunay_triangulation_and_its_dual_2d.Delaunay(points=points)
with cProfile.Profile() as profiler:
    i.compute_voronoi_tessellation()
    profiler_stat = pstats.Stats(profiler).sort_stats(
        pstats.SortKey.CUMULATIVE
    )
    profiler_stat.print_stats()



import numpy as np
from numpy.typing import NDArray
import scipy.spatial


class MockedDelaunay(scipy.spatial.Delaunay):
    def __init__(
        self, points: NDArray[np.float32], simplices: NDArray[np.intc]
    ) -> None:
        self._points = points
        self.simplices = simplices


class MockedVoronoi(scipy.spatial.Delaunay):
    def __init__(
        self,
        points: NDArray[np.float32],
        vertices: NDArray[np.float32],
        ridge_points: NDArray[np.int32],
        ridge_vertices: list[list[int]],
        regions: list[list[int]],
        furthest_site: bool = False,
    ) -> None:
        self._points = points
        self.vertices = vertices
        self.ridge_points = ridge_points
        self.ridge_vertices = ridge_vertices
        self.regions = regions
        self.furthest_site = furthest_site



from abc import ABC, abstractmethod
from collections import defaultdict, deque
import logging
from typing import Literal

import matplotlib.pyplot as plt
import numpy as np
from numpy.typing import NDArray
import scipy
from scipy.spatial import voronoi_plot_2d

from . import exceptions, util
from .geometry import BoundingBox2d, CentroidHelper, CircumcircleHelper


logger = logging.getLogger(__name__)


class Base(ABC):
    def __init__(self, *, points: NDArray[np.float_]) -> None:
        """

        Args:
            points:
                A collection of "n" 2D points arranged in the Cartesian
                coordinate system with the axis order (x, y) and having a shape
                of (n, 2).
        """
        assert self._check_points_validity(points=points)

        self._points: NDArray[np.float_] = points
        self._delaunay_triangulation_computed: bool = False
        self._ridges_and_regions_computed: bool = False
        self._voronoi_tessellation_computed: bool = False
        self._barycentric_dual_computed: bool = False

        self._bounding_box: BoundingBox2d = self._compute_bounding_box(
            points=points
        )

        self._triangles: NDArray[np.intc]
        self._ridge_points: NDArray[np.int32]
        self._ridge_vertices: list[list[int]]
        self._regions: list[list[int]]
        self._voronoi_vertices: NDArray[np.float_]
        self._barycentric_dual_vertices: NDArray[np.float_]

        self._delaunay_edge_to_triangle_indices_mapping: dict[
            tuple[int, int], list[int]
        ]
        self._dual_edges_to_delaunay_edges_mapping: dict[
            tuple[int, int], tuple[int, int]
        ]

    @property
    def points(self) -> NDArray[np.float_]:
        return self._points

    @property
    def bounding_box(self) -> BoundingBox2d:
        return self._bounding_box

    @property
    def triangles(self) -> NDArray[np.intc]:
        self.compute_delaunay_triangulation()
        return self._triangles

    @property
    def ridge_points(self) -> NDArray[np.int32]:
        self._compute_ridges_and_regions()
        return self._ridge_points

    @property
    def ridge_vertices(self) -> list[list[int]]:
        self._compute_ridges_and_regions()
        return self._ridge_vertices

    @property
    def regions(self) -> list[list[int]]:
        self._compute_ridges_and_regions()
        return self._regions

    @property
    def voronoi_vertices(self) -> NDArray[np.float_]:
        self.compute_voronoi_tessellation()
        return self._voronoi_vertices

    @property
    def barycentric_dual_vertices(self) -> NDArray[np.float_]:
        self.compute_barycentric_dual()
        return self._barycentric_dual_vertices

    @property
    def dual_edges_to_delaunay_edges_mapping(
        self,
    ) -> dict[tuple[int, int], tuple[int, int]]:
        self._compute_dual_edges_to_delaunay_edges_mapping()
        return self._dual_edges_to_delaunay_edges_mapping

    @staticmethod
    def _check_points_validity(points: NDArray[np.float_]) -> bool:
        return (
            points.ndim == 2 and points.shape[0] > 2 and points.shape[1] == 2
        )

    @staticmethod
    def _compute_bounding_box(points: NDArray[np.float_]) -> BoundingBox2d:
        return BoundingBox2d(
            min_=np.min(points, axis=0), max_=np.max(points, axis=0)
        )

    @staticmethod
    def _duplicated_points_exist(points: NDArray[np.float32]) -> bool:
        _, count = np.unique(points, axis=0, return_counts=True)
        return np.any(count > 1).item()

    @staticmethod
    def _plot_voronoi_diagram(
        voronoi_diagram: scipy.spatial._qhull.Voronoi,
        bounding_box: BoundingBox2d,
    ):
        fig = voronoi_plot_2d(
            voronoi_diagram,
            show_vertices=False,
            line_colors="orange",
            line_alpha=0.6,
            point_size=2,
        )
        ax = fig.axes[0]
        lower_lim, upper_lim = np.min(bounding_box.min_), np.max(
            bounding_box.max_
        )
        ax.set_xlim(lower_lim, upper_lim)
        ax.set_ylim(lower_lim, upper_lim)
        plt.show()

    @classmethod
    def _jitter_points_on_overlap(
        cls,
        points: NDArray[np.float_],
        rng: np.random.Generator,
        scale: float | None = None,
    ) -> NDArray[np.float_]:
        """Apply a slight jittering to all points in case there are points
        with identical coordinates.

        This process guarantees that no two points share the same coordinates.

        Args:
            points:
                See `__init__`.
            rng:
                numpy random number generator.
            scale:
                The magnitude of movement used for jittering the points. If
                set to None, the scale is automatically determined based on
                the bounding box size of the points.

        Returns:
            Original points if no jittering is made. Otherwise, return the
            jittered points.
        """
        bounding_box = cls._compute_bounding_box(points=points)
        if np.isclose(bounding_box.area, 0.0):
            raise ValueError("Points must span a non-zero area")
        if scale is None:
            scale = (
                np.max(bounding_box.max_) - np.min(bounding_box.min_)
            ) * 1e-6
        while cls._duplicated_points_exist(points=points):
            logger.debug("Jitter points as identical points are found")
            is_addition = (
                points
                <= bounding_box.min_
                + (bounding_box.max_ - bounding_box.min_) / 2
            )
            offset = rng.random(size=points.shape) * scale
            points = np.where(is_addition, points + offset, points - offset)
        return points

    @staticmethod
    def _compute_centroid_of_vertices(
        polygon: NDArray[np.float_],
    ) -> NDArray[np.float_]:
        """Find the centroid of vertices.

        The formula used by this method is given here: https://en.wikipedia.org/wiki/Centroid#Of_a_polygon

        Args:
            polygon:
                NDArray of shape (n, 2) for the `n` vertices of the polygon.

        Returns:
            NDArray of shape (2,) representing the x, y coordinates of the
            centroid.
        """
        if len(polygon) <= 3:
            raise ValueError("Polygon must have at least 3 vertices.")
        area = 0
        centroid = np.zeros((2,), dtype=polygon.dtype)
        for i in range(len(polygon) - 1):
            step = (polygon[i, 0] * polygon[i + 1, 1]) - (
                polygon[i + 1, 0] * polygon[i, 1]
            )
            area += step
            centroid[0] += (polygon[i, 0] + polygon[i + 1, 0]) * step
            centroid[1] += (polygon[i, 1] + polygon[i + 1, 1]) * step
        area /= 2
        centroid *= 1.0 / (6.0 * area)
        return centroid

    @classmethod
    def _compute_centroids_of_voronoi_diagram(
        cls,
        voronoi_diagram: scipy.spatial._qhull.Voronoi,
        target_regions_indices: list[int] | None = None,
    ) -> NDArray[np.float_]:
        centroids: list[NDArray[np.float_]] = []
        if target_regions_indices is None:
            target_regions_indices = voronoi_diagram.point_region
        target_regions = [
            voronoi_diagram.regions[i] for i in target_regions_indices
        ]
        for region in target_regions:
            if len(region) == 0 or -1 in region:
                raise ValueError("Unexpected value in region")
            # Enclose the polygon
            polygon_vertices_indices = region + [region[0]]
            # Get the polygon (represented by a list of vertices coordinates)
            # of this region
            polygon = voronoi_diagram.vertices[polygon_vertices_indices]
            # Find the centroid of those vertices
            centroids.append(
                cls._compute_centroid_of_vertices(polygon=polygon)
            )
        return np.array(centroids)

    @abstractmethod
    def compute_delaunay_triangulation(self) -> None:
        pass

    @abstractmethod
    def compute_voronoi_tessellation(self) -> None:
        pass

    @abstractmethod
    def compute_barycentric_dual(self) -> None:
        pass

    @abstractmethod
    def _compute_ridges_and_regions(self) -> None:
        pass

    @abstractmethod
    def _compute_dual_edges_to_delaunay_edges_mapping(self) -> None:
        pass

    # @staticmethod
    # def _are_line_segments_within_rectangle(
    #     min_x: float,
    #     min_y: float,
    #     max_x: float,
    #     max_y: float,
    #     line_segments: NDArray[np.float_],
    # ) -> NDArray[np.bool_]:
    #     """
    #     Args:
    #         line_segments: numpy array of shape (N, 2, 2), where N is the
    #         number of line segments, the 1st "2" denotes the start and end
    #         points of the each line segment and the 2nd "2" denotes the
    #         (x, y) coordinates of the points.
    #     """
    #     bounds = np.array([[min_x, min_y], [max_x, max_y]])
    #     return np.all(
    #         (line_segments >= bounds[0]) & (line_segments <= bounds[1]),
    #         axis=(1, 2),
    #     )

    def _compute_bounded_line_segments_of_dual(
        self, dual: Literal["voronoi", "barycentric"]
    ) -> NDArray[np.float_]:
        if dual == "voronoi":
            dual_vertices = self.voronoi_vertices
        elif dual == "barycentric":
            dual_vertices = self.barycentric_dual_vertices
        else:
            raise ValueError(
                f"Unexpected value for dual: {dual}. Expected either 'voronoi' or 'barycentric'."
            )
        self._compute_ridges_and_regions()
        ridge_vertices = np.array(self._ridge_vertices, dtype=np.int_)
        mask = ridge_vertices[:, 1] == -1
        target_indices = ridge_vertices[mask, 0]
        target_vertices = dual_vertices[target_indices]
        num_of_vertices = len(target_vertices)
        min_x = self._bounding_box.min_x
        min_y = self._bounding_box.min_y
        max_x = self._bounding_box.max_x
        max_y = self._bounding_box.max_y
        candidates = np.empty((num_of_vertices, 4, 2), dtype=np.float_)
        candidates[:, 0] = np.column_stack(
            (target_vertices[:, 0], np.full(num_of_vertices, min_y))
        )
        candidates[:, 1] = np.column_stack(
            (target_vertices[:, 0], np.full(num_of_vertices, max_y))
        )
        candidates[:, 2] = np.column_stack(
            (np.full(num_of_vertices, min_x), target_vertices[:, 1])
        )
        candidates[:, 3] = np.column_stack(
            (np.full(num_of_vertices, max_x), target_vertices[:, 1])
        )
        distances_between_targets_and_candidates = np.linalg.norm(
            np.expand_dims(target_vertices, axis=1) - candidates, axis=2
        )
        chosen_candidates_index = np.argmin(
            distances_between_targets_and_candidates, axis=1
        )
        chosen_candidates = candidates[
            np.arange(num_of_vertices), chosen_candidates_index, :
        ]
        unbounded_line_segments = np.empty(
            (len(ridge_vertices), 2, 2), dtype=np.float_
        )
        unbounded_line_segments[mask] = np.stack(
            (target_vertices, chosen_candidates), axis=1
        )
        unbounded_line_segments[~mask] = dual_vertices[ridge_vertices[~mask]]
        bounded_line_segments = self._bounding_box.clip_line_segments(
            line_segments=unbounded_line_segments
        )
        return bounded_line_segments

    def get_mocked_scipy_spatial_delaunay(self) -> util.MockedDelaunay:
        """Get an object to be used as the argument in
        `scipy.spatial.delaunay_plot_2d`.
        """
        self.compute_delaunay_triangulation()
        return util.MockedDelaunay(
            points=self._points, simplices=self._triangles
        )

    def get_mocked_scipy_spatial_voronoi(self) -> util.MockedVoronoi:
        """Get an object to be used as the argument in
        `scipy.spatial.voronoi_plot_2d`.
        """
        self.compute_voronoi_tessellation()
        return util.MockedVoronoi(
            points=self._points,
            vertices=self._voronoi_vertices,
            ridge_points=self._ridge_points,
            ridge_vertices=self._ridge_vertices,
            regions=self._regions,
        )

    def get_barycentric_dual_as_mocked_scipy_spatial_voronoi(
        self,
    ) -> util.MockedVoronoi:
        """Get an object to be used as the argument in
        `scipy.spatial.voronoi_plot_2d`.
        """
        self.compute_barycentric_dual()
        return util.MockedVoronoi(
            points=self._points,
            vertices=self._barycentric_dual_vertices,
            ridge_points=self._ridge_points,
            ridge_vertices=self._ridge_vertices,
            regions=self._regions,
        )

    @classmethod
    def voronoi_iteration(
        cls,
        points: NDArray[np.float_],
        rng: np.random.Generator,
        debug: bool = False,
    ) -> NDArray[np.float_]:
        assert cls._check_points_validity(points=points)

        # Make sure there is no overlapping points as it would result in a
        # reduced number of regions compared to the initial number of input
        # points after the Voronoi iteration.
        points = cls._jitter_points_on_overlap(points=points, rng=rng)
        bounding_box = cls._compute_bounding_box(points=points)
        dilated_bounding_box = bounding_box.get_dilated_bounding_box(
            dilation=bounding_box.dimensions
        )
        dilated_bounding_box_vertices = dilated_bounding_box.get_vertices()
        points_with_dilated_bounding_box_vertices = np.append(
            arr=points,
            values=dilated_bounding_box_vertices,
            axis=0,
        )
        # NOTE: Not sure whether qhull_options should be changed
        voronoi_diagram = scipy.spatial.Voronoi(
            points_with_dilated_bounding_box_vertices,
            # qhull_options="Qbb Qc Qx",
        )
        if __debug__:
            if debug:
                cls._plot_voronoi_diagram(voronoi_diagram=voronoi_diagram)

        new_points = cls._compute_centroids_of_voronoi_diagram(
            voronoi_diagram=voronoi_diagram,
            target_regions_indices=voronoi_diagram.point_region[:-4],
        )
        new_points = np.clip(
            new_points, a_min=bounding_box.min_, a_max=bounding_box.max_
        )
        return new_points


class Delaunay(Base):
    def __init__(self, *, points: NDArray[np.float_]) -> None:
        super().__init__(points=points)

        self.compute_delaunay_triangulation()

    def _compute_ridges_and_regions(self) -> None:
        if self._ridges_and_regions_computed:
            return
        self.compute_delaunay_triangulation()
        triangles = self._triangles
        # Retrieve all the edges from triangles
        delaunay_edges = np.moveaxis(
            np.array(
                [
                    [triangles[:, 0], triangles[:, 1]],
                    [triangles[:, 1], triangles[:, 2]],
                    [triangles[:, 2], triangles[:, 0]],
                ]
            ),
            source=2,
            destination=0,
        )
        # Sort the vertices indices of each edge
        delaunay_edges.sort(axis=2)
        # Convert delaunay_edges to Tuple
        delaunay_edges = tuple(
            tuple(map(tuple, edges)) for edges in delaunay_edges
        )
        # Build a mapping from edge to triangles
        delaunay_edge_to_triangle_indices_mapping = defaultdict(list)
        for triangle_index, triangle_edges in enumerate(delaunay_edges):
            for edge in triangle_edges:
                delaunay_edge_to_triangle_indices_mapping[edge].append(
                    triangle_index
                )

        regions: list[list[int]] = [[] for _ in range(len(self._points))]
        indices_of_visited_triangles: set[int] = set()
        dfs_stack = deque()
        dfs_stack.append(0)
        while len(dfs_stack):
            triangle_index = dfs_stack.pop()
            if triangle_index in indices_of_visited_triangles:
                continue
            indices_of_visited_triangles.add(triangle_index)
            triangle = triangles[triangle_index]
            for point_index in triangle:
                regions[point_index].append(triangle_index)
            for edge in delaunay_edges[triangle_index]:
                for (
                    triangle_index
                ) in delaunay_edge_to_triangle_indices_mapping[edge]:
                    dfs_stack.append(triangle_index)

        def connect_triangles(triangles_indices: list[int]) -> list[int]:
            # Build a mapping from edge to triangles
            delaunay_edge_to_triangle_indices_mapping = defaultdict(list)
            for triangle_index in triangles_indices:
                for edge in delaunay_edges[triangle_index]:
                    delaunay_edge_to_triangle_indices_mapping[edge].append(
                        triangle_index
                    )
            # Connect the triangles
            result_reversed: bool = False
            result: list[int] = []
            result.append(triangles_indices[0])
            while len(result) != len(triangles_indices):
                common_edge_found = False
                for edge in delaunay_edges[result[-1]]:
                    for (
                        affiliated_triangle_index
                    ) in delaunay_edge_to_triangle_indices_mapping[edge]:
                        if affiliated_triangle_index not in result:
                            result.append(affiliated_triangle_index)
                            common_edge_found = True
                            break
                    if common_edge_found:
                        break
                if common_edge_found is False:
                    assert not result_reversed
                    result.reverse()
                    result_reversed = True
            # Check whether the head and the tail are connected
            # If not connected, append -1 to the result
            head_edges = set(edge for edge in delaunay_edges[result[0]])
            tail_edges = set(edge for edge in delaunay_edges[result[-1]])
            if len(head_edges.intersection(tail_edges)) == 0:
                result.append(-1)
            return result

        for index, region in enumerate(regions):
            if len(region) < 3:
                regions[index].append(-1)
                continue
            regions[index] = connect_triangles(triangles_indices=region)

        ridge_points = []
        ridge_vertices = []
        for (
            delaunay_edge,
            triangle_indices,
        ) in delaunay_edge_to_triangle_indices_mapping.items():
            assert len(triangle_indices) == 1 or len(triangle_indices) == 2
            ridge_points.append(delaunay_edge)
            ridge_vertices.append(
                triangle_indices
                if len(triangle_indices) == 2
                else [triangle_indices[0], -1]
            )
        ridge_points = np.array(ridge_points)

        self._ridge_points = ridge_points
        self._ridge_vertices = ridge_vertices
        self._regions = regions
        self._delaunay_edge_to_triangle_indices_mapping = dict(
            delaunay_edge_to_triangle_indices_mapping
        )
        self._ridges_and_regions_computed = True

    def _compute_dual_edges_to_delaunay_edges_mapping(self) -> None:
        self._compute_ridges_and_regions()
        dual_edges_to_delaunay_edges_mapping = {}
        for (
            delaunay_edge,
            triangle_indices,
        ) in self._delaunay_edge_to_triangle_indices_mapping.items():
            if len(triangle_indices) == 1:
                continue
            assert len(triangle_indices) == 2
            dual_edges_to_delaunay_edges_mapping[
                (triangle_indices[0], triangle_indices[1])
            ] = delaunay_edge
        self._dual_edges_to_delaunay_edges_mapping = (
            dual_edges_to_delaunay_edges_mapping
        )

    def compute_delaunay_triangulation(self) -> None:
        if self._delaunay_triangulation_computed:
            return
        delaunay_graph = scipy.spatial.Delaunay(points=self._points)
        triangles = delaunay_graph.simplices
        self._triangles = triangles
        self._delaunay_triangulation_computed = True

    def compute_voronoi_tessellation(self) -> None:
        if self._voronoi_tessellation_computed:
            return
        self._compute_ridges_and_regions()
        # Compute the circumcenter for each triangle, which is a site in
        # voronoi diagram
        circumcenters = CircumcircleHelper.compute_circumcenters(
            points=self._points[self._triangles]
        )

        # Check for duplicated circumcenters
        unique_circumcenters = np.unique(
            circumcenters,
            axis=0,
            return_counts=False,
            return_index=False,
            return_inverse=False,
        )
        if len(unique_circumcenters) != len(circumcenters):
            raise exceptions.DuplicatedCircumcentersError(
                "Cannot compute voronoi diagram due to duplicated circumcenters"
            )

        self._voronoi_vertices = circumcenters
        self._voronoi_tessellation_computed = True

    def compute_barycentric_dual(self) -> None:
        if self._barycentric_dual_computed:
            return
        self._compute_ridges_and_regions()
        # Compute the centroid for each triangle
        triangles = self._triangles
        centroids = CentroidHelper.compute_centroids(
            points=self._points[triangles]
        )

        self._barycentric_dual_vertices = centroids
        self._barycentric_dual_computed = True



import numpy as np
from numpy.typing import NDArray


class CentroidHelper:
    @staticmethod
    def compute_centroids(points: NDArray[np.float32]) -> NDArray[np.float32]:
        return np.mean(points, axis=1)



from __future__ import annotations

import dataclasses

import numpy as np
from numpy.typing import NDArray


@dataclasses.dataclass(frozen=True)
class BoundingBox2d:
    min_: NDArray[np.float_]
    max_: NDArray[np.float_]

    def __post_init__(self) -> None:
        """

        Args:
            min_:
                A 2D point in the Cartesian coordinate system represented by an
                array of shape (2,).
            max_:
                See `min_`.
        """
        assert self.min_.shape == (2,), self.min_.shape
        assert self.max_.shape == (2,), self.max_.shape
        assert np.all(self.min_ <= self.max_)

    @property
    def min_x(self) -> float:
        return self.min_[0]

    @property
    def min_y(self) -> float:
        return self.min_[1]

    @property
    def max_x(self) -> float:
        return self.max_[0]

    @property
    def max_y(self) -> float:
        return self.max_[1]

    @property
    def dimensions(self) -> NDArray[np.float_]:
        return self.max_ - self.min_

    @property
    def area(self) -> float:
        return np.prod(self.dimensions)

    @property
    def midpoint(self) -> NDArray[np.float_]:
        return (self.min_ + self.max_) / 2

    def get_vertices(self) -> NDArray[np.float_]:
        return np.array(
            [
                [self.min_x, self.min_y],
                [self.min_x, self.max_y],
                [self.max_x, self.max_y],
                [self.max_x, self.min_y],
            ]
        )

    def get_dilated_bounding_box(
        self, dilation: NDArray[np.float_] | float
    ) -> BoundingBox2d:
        return BoundingBox2d(
            min_=self.min_ - dilation, max_=self.max_ + dilation
        )

    def liang_barsky_line_clipping(
        self, line_segments: NDArray[np.float_]
    ) -> NDArray[np.float_]:
        x1, y1 = line_segments[:, 0].T
        x2, y2 = line_segments[:, 1].T
        dx = x2 - x1
        dy = y2 - y1
        p = np.stack([-dx, dx, -dy, dy], dtype=np.float_)
        q = np.stack(
            [
                x1 - self.min_x,
                self.max_x - x1,
                y1 - self.min_y,
                self.max_y - y1,
            ],
            dtype=np.float_,
        )
        t_enter = np.zeros_like(x1)
        t_exit = np.ones_like(x1)

        for i in range(4):
            is_parallel = p[i] == 0
            outside_parallel = np.logical_and(is_parallel, q[i] < 0)
            # t = np.where(p[i] != 0, q[i] / p[i], np.inf)
            # To avoid RuntimeWarning: divide by zero encountered in divide
            t = np.divide(
                q[i], p[i], out=np.full_like(q[i], np.inf), where=p[i] != 0
            )
            t_enter = np.where(
                np.logical_and(p[i] < 0, ~outside_parallel),
                np.maximum(t_enter, t),
                t_enter,
            )
            t_exit = np.where(p[i] > 0, np.minimum(t_exit, t), t_exit)

        outside = t_enter > t_exit
        x1_clip = np.where(~outside, x1 + t_enter * dx, np.nan)
        y1_clip = np.where(~outside, y1 + t_enter * dy, np.nan)
        x2_clip = np.where(~outside, x1 + t_exit * dx, np.nan)
        y2_clip = np.where(~outside, y1 + t_exit * dy, np.nan)

        result = np.stack(
            [
                np.column_stack((x1_clip, y1_clip)),
                np.column_stack((x2_clip, y2_clip)),
            ],
            axis=1,
        )
        result = np.where(
            np.all(np.isclose(result[:, 0], result[:, 1]), axis=1),
            np.full_like(result, fill_value=np.nan).T,
            result.T,
        ).T
        return result

    def clip_line_segments(
        self, line_segments: NDArray[np.float_]
    ) -> NDArray[np.float_]:
        return self.liang_barsky_line_clipping(line_segments=line_segments)



import timeit

import matplotlib.pyplot as plt


POINT_COUNT_TEST_CASES = [100, 1000, 10000, 100000, 1000000]
NUM_REPEATS_PER_TEST_CASE = [16, 8, 4, 2, 1]
GENERATE_RANDOM_POINTS_SETUP_CODE = r"""
import numpy as np
import scipy.spatial

import delaunay_triangulation_and_its_dual_2d

rng = np.random.default_rng()
points = rng.uniform(low=0.0, high=1.0, size=({}, 2))
"""
DELAUNAY_CLASS_COMPUTE_VORONOI_TESSELLATION_STATEMENT = r"""
i = delaunay_triangulation_and_its_dual_2d.Delaunay(points=points)
i.compute_voronoi_tessellation()
"""
SCIPY_SPATIAL_VORONOI_CLASS_COMPUTE_VORONOI_TESSELLATION_STATEMENT = (
    r"scipy.spatial.Voronoi(points=points)"
)

if __name__ == "__main__":
    delaunay_class_compute_voronoi_tessellation_time_results = []
    scipy_spatial_voronoi_class_compute_voronoi_tessellation_time_results = []
    for point_count, num_repeats in zip(
        POINT_COUNT_TEST_CASES, NUM_REPEATS_PER_TEST_CASE
    ):
        generate_random_points_setup_code = (
            GENERATE_RANDOM_POINTS_SETUP_CODE.format(point_count)
        )
        delaunay_class_compute_voronoi_tessellation_time = timeit.timeit(
            stmt=DELAUNAY_CLASS_COMPUTE_VORONOI_TESSELLATION_STATEMENT,
            setup=generate_random_points_setup_code,
            number=num_repeats,
        )
        delaunay_class_compute_voronoi_tessellation_time_results.append(
            delaunay_class_compute_voronoi_tessellation_time
        )
        scipy_spatial_voronoi_class_compute_voronoi_tessellation_time = timeit.timeit(
            stmt=SCIPY_SPATIAL_VORONOI_CLASS_COMPUTE_VORONOI_TESSELLATION_STATEMENT,
            setup=generate_random_points_setup_code,
            number=num_repeats,
        )
        scipy_spatial_voronoi_class_compute_voronoi_tessellation_time_results.append(
            scipy_spatial_voronoi_class_compute_voronoi_tessellation_time
        )

    plt.plot(
        POINT_COUNT_TEST_CASES,
        delaunay_class_compute_voronoi_tessellation_time_results,
        label="compute_voronoi_tessellation",
    )
    plt.plot(
        POINT_COUNT_TEST_CASES,
        scipy_spatial_voronoi_class_compute_voronoi_tessellation_time_results,
        label="scipy.spatial.Voronoi",
    )
    plt.xlabel("Num of Points")
    plt.ylabel("Time (s)")
    plt.xscale("log")
    plt.yscale("log")
    plt.title("Time to compute Voronoi tessellation")
    plt.legend()
    plt.grid()
    plt.show()



import sys

import matplotlib.collections
import matplotlib.pyplot as plt
import numpy as np
import scipy.spatial

import delaunay_triangulation_and_its_dual_2d
import delaunay_triangulation_and_its_dual_2d.exceptions


def visualize_mocked_scipy_spatial_delaunay_and_voronoi():
    points = np.array(
        [[0, 0], [3, 0], [1, 1], [3, 2], [4, 2], [1, 3], [2, 4], [4, 4]],
        dtype=np.float32,
    )
    delaunay = delaunay_triangulation_and_its_dual_2d.Delaunay(points=points)
    delaunay_diagram = delaunay.get_mocked_scipy_spatial_delaunay()
    fig = scipy.spatial.delaunay_plot_2d(delaunay_diagram)
    barycentric_dual_diagram = (
        delaunay.get_barycentric_dual_as_mocked_scipy_spatial_voronoi()
    )
    fig = scipy.spatial.voronoi_plot_2d(
        barycentric_dual_diagram, ax=fig.axes[0]
    )
    voronoi_diagram = delaunay.get_mocked_scipy_spatial_voronoi()
    fig = scipy.spatial.voronoi_plot_2d(voronoi_diagram, ax=fig.axes[0])
    fig.axes[0].set_xlim(-2, 6)
    fig.axes[0].set_ylim(-2, 6)
    plt.grid()
    plt.show()


# def visualize_are_line_segments_within_rectangle():
#     line_segments = np.random.randint(0, 10, (5, 2, 2)).astype(np.float_)
#     line_segments_within_rectangle = delaunay_triangulation_and_its_dual_2d.Delaunay._are_line_segments_within_rectangle(
#         min_x=2, min_y=2, max_x=8, max_y=8, line_segments=line_segments
#     )
#     colors = np.where(line_segments_within_rectangle, "green", "red")
#     fig, ax = plt.subplots()
#     # Reference: https://stackoverflow.com/a/21357666
#     ax.add_collection(
#         matplotlib.collections.LineCollection(line_segments, colors=colors)
#     )
#     rectangle = np.array(
#         [
#             [[2, 2], [2, 8]],
#             [[2, 8], [8, 8]],
#             [[8, 8], [8, 2]],
#             [[8, 2], [2, 2]],
#         ]
#     ).astype(np.float_)
#     ax.add_collection(
#         matplotlib.collections.LineCollection(rectangle, colors="blue")
#     )
#     ax.autoscale()
#     ax.grid()
#     plt.show()


def try_compute_bounded_line_segments_of_dual():
    points = np.array(
        [[0, 0], [3, 0], [1, 1], [3, 2], [4, 2], [1, 3], [2, 4], [4, 4]],
        dtype=np.float32,
    )
    delaunay = delaunay_triangulation_and_its_dual_2d.Delaunay(points=points)
    bounded_line_segments = delaunay._compute_bounded_line_segments_of_dual(
        dual="voronoi"
    )
    print("bounded_line_segments.shape: ", bounded_line_segments.shape)


if __name__ == "__main__":
    # Reference: https://stackoverflow.com/a/52837375
    globals()[sys.argv[1]]()



import dataclasses

import numpy as np
from numpy.typing import NDArray


@dataclasses.dataclass
class Circumcircle:
    center: tuple[float, float]
    radius: float


class CircumcircleHelper:
    @staticmethod
    def compute_circumcenters(
        points: NDArray[np.float32],
    ) -> NDArray[np.float32]:
        """Compute the circumcenters of triangles.

        Args:
            points:
                An array of shape (n, 3, 2). The first dimension "n"
                refers to the number of batches. The second dimension refers to
                the three vertices of a triangle used for calculating the
                circumcenter. The third dimension represents the (x, y)
                Cartesian coordinates of the points.
        """
        return CircumcircleHelper.compute_circumcenters_v4(points=points)

    @staticmethod
    def compute_circumcenters_v3(
        points: NDArray[np.float32],
    ) -> NDArray[np.float32]:  # pragma: no cover
        """Compute the circumcenters of triangles.

        Args:
            points:
                See method `compute_circumcenters`.
        """
        midpoints = (points[:, :2] + points[:, 1:]) / 2
        slopes = (points[:, 1:, 1] - points[:, :2, 1]) / (
            points[:, 1:, 0] - points[:, :2, 0]
        )
        orthogonal_slopes = -1 / slopes

        # Solve Ax = b for the circumcenter
        matrix_a = np.stack(
            [orthogonal_slopes, -np.ones_like(orthogonal_slopes)], axis=-1
        )
        b = np.sum(midpoints * matrix_a, axis=-1)
        circumcenter = np.linalg.solve(matrix_a, b)
        return circumcenter

    @staticmethod
    def compute_circumcenters_v4(
        points: NDArray[np.float32],
    ) -> NDArray[np.float32]:
        """Compute the circumcenters of triangles.

        Args:
            points:
                See method `compute_circumcenters`.

        References:
            - How do I find the circumcenter of the triangle using python
                without external libraries?
                [link](https://stackoverflow.com/a/56225021)
        """
        A, B, C = points[:, 0], points[:, 1], points[:, 2]
        Ax, Ay = A[:, 0], A[:, 1]
        Bx, By = B[:, 0], B[:, 1]
        Cx, Cy = C[:, 0], C[:, 1]

        D = 2 * (Ax * (By - Cy) + Bx * (Cy - Ay) + Cx * (Ay - By))
        Ux = (
            (Ax**2 + Ay**2) * (By - Cy)
            + (Bx**2 + By**2) * (Cy - Ay)
            + (Cx**2 + Cy**2) * (Ay - By)
        ) / D
        Uy = (
            (Ax**2 + Ay**2) * (Cx - Bx)
            + (Bx**2 + By**2) * (Ax - Cx)
            + (Cx**2 + Cy**2) * (Bx - Ax)
        ) / D

        return np.stack((Ux, Uy), axis=-1)
